"""Utility functions for the Hysetter package."""

from __future__ import annotations

import logging
from logging import Logger

from rich.logging import RichHandler

__all__ = ["get_logger"]


def get_logger() -> Logger:
    """Set up Rich Logger and return it."""
    logging.basicConfig(
        level="INFO",
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(rich_tracebacks=True)],
    )

    return logging.getLogger("hysetter")



"""Customized PyNHD exceptions."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence


class DataNotAvailableError(Exception):
    """Exception raised for requested data is not available.

    Parameters
    ----------
    data_name : str
        Data name requested.
    """

    def __init__(self, data_name: str) -> None:
        self.message = f"{data_name.capitalize()} is not available for the requested query."
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class EmptyResponseError(Exception):
    """Exception raised when the input response is empty."""

    def __init__(self) -> None:
        self.message = "The input response is empty."
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class ServiceUnavailableError(Exception):
    """Exception raised when the service is not available.

    Parameters
    ----------
    url : str
        The server url
    """

    def __init__(self, url: str) -> None:
        self.message = f"Service is currently not available, try again later:\n{url}"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class ZeroMatchedError(ValueError):
    """Exception raised when a function argument is missing.

    Parameters
    ----------
    msg : str
        The exception error message
    """

    def __init__(self, msg: str | None = None) -> None:
        if msg is None:
            self.message = "Service returned no features."
        else:
            self.message = f"Service returned no features with the following error message:\n{msg}"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class MissingColumnError(Exception):
    """Exception raised when a required column is missing from a dataframe.

    Parameters
    ----------
    missing : list
        List of missing columns.
    """

    def __init__(self, missing: list[str]) -> None:
        self.message = f"The following columns are missing:\n{', '.join(missing)}"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class ServiceError(Exception):
    """Exception raised when the requested data is not available on the server.

    Parameters
    ----------
    err : str
        Service error message.
    """

    def __init__(self, err: str, url: str | None = None) -> None:
        self.message = "Service returned the following error message:\n"
        if url is None:
            self.message += err
        else:
            self.message += f"URL: {url}\nERROR: {err}\n"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class InputValueError(Exception):
    """Exception raised for invalid input.

    Parameters
    ----------
    inp : str
        Name of the input parameter
    valid_inputs : tuple
        List of valid inputs
    given : str, optional
        The given input, defaults to None.
    """

    def __init__(
        self,
        inp: str,
        valid_inputs: Sequence[str | int] | Generator[str | int, None, None],
        given: str | int | None = None,
    ) -> None:
        if given is None:
            self.message = f"Given {inp} is invalid. Valid options are:\n"
        else:
            self.message = f"Given {inp} ({given}) is invalid. Valid options are:\n"
        self.message += "\n".join(str(i) for i in valid_inputs)
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class InputTypeError(Exception):
    """Exception raised when a function argument type is invalid.

    Parameters
    ----------
    arg : str
        Name of the function argument
    valid_type : str
        The valid type of the argument
    example : str, optional
        An example of a valid form of the argument, defaults to None.
    """

    def __init__(
        self,
        arg: str,
        valid_type: str,
        example: str | None = None,
    ) -> None:
        self.message = f"The {arg} argument should be of type {valid_type}"
        if example is not None:
            self.message += f":\n{example}"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class MissingItemError(Exception):
    """Exception raised when a required item is missing.

    Parameters
    ----------
    missing : tuple
        The missing items.
    """

    def __init__(self, missing: list[str]) -> None:
        self.message = f"The following items are missing:\n{', '.join(missing)}"
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class InputRangeError(Exception):
    """Exception raised when a function argument is not in the valid range.

    Parameters
    ----------
    variable : str
        Variable with invalid value
    valid_range : str
        Valid range
    """

    def __init__(self, variable: str, valid_range: str) -> None:
        self.message = f"Valid range for {variable} is {valid_range}."
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class MissingCRSError(Exception):
    """Exception raised when CRS is not given."""

    def __init__(self) -> None:
        self.message = "CRS of the input geometry is missing."
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class NoTerminalError(Exception):
    """Exception raised when no terminal COMID is found."""

    def __init__(self) -> None:
        self.message = "No terminal COMID was found."
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message


class DependencyError(Exception):
    """Exception raised when a dependencies are not met.

    Parameters
    ----------
    libraries : tuple
        List of valid inputs
    """

    def __init__(self, func: str, libraries: str | list[str] | Generator[str, None, None]) -> None:
        libraries = [libraries] if isinstance(libraries, str) else libraries
        self.message = f"The following dependencies are missing for running {func}:\n"
        self.message += ", ".join(libraries)
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the error message."""
        return self.message



from __future__ import annotations

import io

import pytest
from click.testing import CliRunner, Result

import hysetter as hs
from hysetter.cli import cli


@pytest.fixture
def runner():
    """Return a CliRunner."""
    return CliRunner()


@pytest.fixture
def run_config_once(runner: CliRunner) -> Result:
    ret = runner.invoke(cli, ["config_demo.yml"])
    return ret


def test_config(run_config_once: Result) -> None:
    assert run_config_once.exit_code == 0


def test_config_overwrite(runner: CliRunner, run_config_once: Result) -> None:
    ret = runner.invoke(cli, ["config_demo.yml", "--overwrite"])
    assert ret.exit_code == 0


def test_show_versions() -> None:
    f = io.StringIO()
    hs.show_versions(file=f)
    assert "SYS INFO" in f.getvalue()



"""Main functions of hysetter."""

from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
from rich.console import Console
from rich.progress import track

if TYPE_CHECKING:
    from hysetter.hysetter import Config, Forcing

__all__ = ["get_forcing"]


def get_forcing(data_cfg: Forcing, model_cfg: Config) -> None:
    """Get forcing for the area of interest."""
    from shapely import box

    console = Console(force_jupyter=False)
    if data_cfg.source == "daymet":
        import pydaymet as daymet

        get_clm = daymet.get_bygeom
    elif data_cfg.source == "gridmet":
        import pygridmet as gridmet

        get_clm = gridmet.get_bygeom
    elif data_cfg.source == "nldas2":
        import pynldas2 as nldas2

        get_clm = nldas2.get_bygeom
    else:
        raise ValueError("Unknown forcing source.")

    gdf = gpd.read_parquet(model_cfg.file_paths.aoi_parquet)
    forcing_dir = model_cfg.file_paths.forcing
    forcing_dir.mkdir()
    forcing_name = {
        "daymet": "Daymet",
        "gridmet": "GridMet",
        "nldas2": "NLDAS2",
    }[data_cfg.source]

    join_style = "round"
    if not data_cfg.crop:
        gdf["geometry"] = box(*gdf.geometry.bounds.to_numpy().T)
        join_style = "mitre"
    if data_cfg.geometry_buffer > 0:
        gdf = gdf.to_crs(5070).buffer(data_cfg.geometry_buffer, join_style=join_style)
    for i, geom in track(
        enumerate(gdf.geometry),
        description=f"Getting forcing from {forcing_name}",
        total=len(gdf),
        console=console,
    ):
        forcing_dir[i] = (
            f"{data_cfg.source}_{data_cfg.start_date.date()}_{data_cfg.end_date.date()}_geom_{i}.nc"
        )
        if forcing_dir[i].exists():
            continue
        try:
            clm = get_clm(
                geom,  # pyright: ignore[reportArgumentType]
                (data_cfg.start_date, data_cfg.end_date),  # pyright: ignore[reportArgumentType]
                gdf.crs,  # pyright: ignore[reportArgumentType]
                data_cfg.variables,  # pyright: ignore[reportArgumentType]
            )
            clm.to_netcdf(forcing_dir[i])
        except Exception:
            console.print_exception(show_locals=True, max_frames=4)
            console.print(f"Failed to get forcing for AOI index {i}.")
            continue



"""Main functions of hysetter."""

from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
from rich.console import Console
from rich.progress import track

if TYPE_CHECKING:
    from hysetter.hysetter import NLCD, Config

__all__ = ["get_nlcd"]


def get_nlcd(data_cfg: NLCD, model_cfg: Config) -> None:
    """Get NLCD data for the area of interest."""
    from pygeohydro import nlcd_bygeom
    from shapely import box

    console = Console(force_jupyter=False)
    gdf = gpd.read_parquet(model_cfg.file_paths.aoi_parquet)
    nlcd_paths = model_cfg.file_paths.nlcd
    nlcd_paths.mkdir()
    years = {
        "cover": data_cfg.cover,
        "impervious": data_cfg.impervious,
        "canopy": data_cfg.canopy,
        "descriptor": data_cfg.descriptor,
    }
    years = {k: v for k, v in years.items() if v is not None}
    join_style = "round"
    if not data_cfg.crop:
        gdf["geometry"] = box(*gdf.geometry.bounds.to_numpy().T)
        join_style = "mitre"
    if data_cfg.geometry_buffer > 0:
        gdf = gdf.to_crs(5070).buffer(data_cfg.geometry_buffer, join_style=join_style)
    if not years:
        years = None
    for i, geom in track(
        enumerate(gdf.geometry),
        description="Getting NLCD from MRLC",
        total=len(gdf),
        console=console,
    ):
        nlcd_paths[i] = f"nlcd_geom_{i}.nc"
        if nlcd_paths[i].exists():
            continue
        try:
            nlcd = nlcd_bygeom(
                gpd.GeoSeries(geom, crs=gdf.crs),  # pyright: ignore[reportCallIssue]
                30,
                years=years,
            )
        except Exception:
            console.print_exception(show_locals=True, max_frames=4)
            console.print(f"Failed to get NLCD data for AOI index {i}")
            continue
        nlcd[0].to_netcdf(nlcd_paths[i])



"""Configuration for pytest."""

from __future__ import annotations

from typing import Any

import pytest


@pytest.fixture(autouse=True)
def _add_standard_imports(doctest_namespace: dict[str, Any]):
    """Add hysetter namespace for doctest."""
    import hysetter as hs

    doctest_namespace["hs"] = hs



"""Command-line interface for hysetter."""

from __future__ import annotations

import importlib.metadata
import shutil
import sys
from pathlib import Path

import rich_click as click
from rich import box
from rich.console import Console
from rich.table import Table

from hysetter import hysetter as hs

console = Console(force_jupyter=False)
CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


def get_versions() -> dict[str, str]:
    """Get the versions of the dependencies."""

    def _get_version(module: str) -> str:
        """Get the version of the module."""
        try:
            return importlib.metadata.version(module)
        except importlib.metadata.PackageNotFoundError:
            return "not installed"

    hs_ver = _get_version("hysetter")
    if "dev" in hs_ver:
        hs_ver = ".".join(hs_ver.split(".")[:3]) + "-dev"
    return {
        "HySetter": hs_ver,
        "HyRiver Stack": ".".join(_get_version("pygeohydro").split(".")[:2]),
        "Python": ".".join(map(str, sys.version_info[:3])),
    }


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("config_yml", type=click.Path(exists=True))
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    help="Overwrite the existing data, defaults to ``False``.",
)
@click.version_option(importlib.metadata.version("hysetter"), prog_name="HySetter")
def cli(config_yml: str, overwrite: bool = False) -> None:
    r"""Get hydroclimate data for areas of interest.

    \b
    CONFIG_YML: Path to the configuration file in YAML format.

    \b
    $ hysetter config.yml --overwrite
    """
    table = Table(box=box.ROUNDED)
    table.add_column("Package", style="bold")
    table.add_column("Version", style="magenta", justify="center")
    for p, v in get_versions().items():
        table.add_row(p, v)
    console.print(table)

    console.print(f"Reading configuration file: [bold green]{Path(config_yml).resolve()}")
    cfg = hs.read_config(config_yml)

    if overwrite:
        console.print("Removing existing data")
        project_dir = Path(cfg.project.data_dir, cfg.project.name.replace(" ", "_"))
        shutil.rmtree(project_dir, ignore_errors=True)
    try:
        cfg.get_data()
    except Exception:
        console.print_exception(extra_lines=3, show_locals=True)
        sys.exit(1)



"""Main functions of hysetter."""

from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
from rich.console import Console
from rich.progress import track

if TYPE_CHECKING:
    from hysetter.hysetter import Config, Soil

__all__ = ["get_soil"]


def get_soil(data_cfg: Soil, model_cfg: Config) -> None:
    """Get soil data for the area of interest."""
    import pygeohydro as gh
    from shapely import box

    console = Console(force_jupyter=False)

    if data_cfg.source == "gnatsgo":
        soil_func = gh.soil_gnatsgo
    elif data_cfg.source == "soilgrids":
        soil_func = gh.soil_soilgrids
    elif data_cfg.source == "polaris":
        soil_func = gh.soil_polaris
    else:
        raise ValueError("Unknown forcing source.")

    gdf = gpd.read_parquet(model_cfg.file_paths.aoi_parquet)
    source_name = {
        "gnatsgo": "gNATSGO",
        "soilgrids": "SoilGrids",
        "polaris": "POLARIS",
    }[data_cfg.source]
    soil_paths = model_cfg.file_paths.soil
    soil_paths.mkdir()
    join_style = "round"
    if not data_cfg.crop:
        gdf["geometry"] = box(*gdf.geometry.bounds.to_numpy().T)
        join_style = "mitre"
    if data_cfg.geometry_buffer > 0:
        gdf = gdf.to_crs(5070).buffer(data_cfg.geometry_buffer, join_style=join_style)
    for i, geom in track(
        enumerate(gdf.geometry),
        description=f"Getting soil from {source_name}",
        total=len(gdf),
        console=console,
    ):
        soil_paths[i] = f"{data_cfg.source}_geom_{i}.nc"
        if soil_paths[i].exists():
            continue
        try:
            soil = soil_func(data_cfg.variables, geom, gdf.crs)  # pyright: ignore[reportArgumentType]
        except Exception:
            console.print_exception(show_locals=True, max_frames=4)
            console.print(f"Failed to get soil for AOI index {i}")
            continue
        soil.to_netcdf(soil_paths[i])



"""Top-level package for hysetter."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

from hysetter import exceptions
from hysetter.aoi import get_aoi
from hysetter.forcing import get_forcing
from hysetter.hysetter import Config, read_config, write_config
from hysetter.nid import get_nid
from hysetter.nlcd import get_nlcd
from hysetter.nwis import get_streamflow
from hysetter.print_versions import show_versions
from hysetter.soil import get_soil
from hysetter.topo import get_topo

try:
    __version__ = version("hysetter")
except PackageNotFoundError:
    __version__ = "999"

__all__ = [
    "Config",
    "__version__",
    "exceptions",
    "get_aoi",
    "get_forcing",
    "get_nid",
    "get_nlcd",
    "get_soil",
    "get_streamflow",
    "get_topo",
    "read_config",
    "show_versions",
    "write_config",
]



"""Module for computing runoff using the NRCS-PDM model with snowmelt."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, overload

import numpy as np
import xarray as xr
from numba import config as nbconfig
from numba import njit, prange

if TYPE_CHECKING:
    from numpy.typing import NDArray

    FloatArray = NDArray[np.floating[Any]]
    UIntArray = NDArray[np.uint32]

nbconfig.THREADING_LAYER = "workqueue"

__all__ = ["compute_kge", "separate_snow", "to_numpy"]

SMALL = 1e-6


@njit("f8(f8[::1], f8[::1])", nogil=True)
def compute_kge(sim: FloatArray, obs: FloatArray) -> float:
    """Compute the original Kling-Gupta Efficiency (2009)."""
    cc = np.corrcoef(obs, sim)[0, 1]
    alpha = np.std(sim) / np.std(obs)
    beta = np.sum(sim) / np.sum(obs)
    kge = 1 - np.sqrt(np.square(cc - 1) + np.square(alpha - 1) + np.square(beta - 1))
    return kge


@overload
def to_numpy(data: Any, dtype: Literal["f8"] = "f8") -> FloatArray: ...


@overload
def to_numpy(data: Any, dtype: Literal["uint32"]) -> UIntArray: ...


def to_numpy(data: Any, dtype: Literal["f8", "uint32"] = "f8") -> FloatArray | UIntArray:
    """Convert data to a contiguous numpy array with specified dtype."""
    return np.asarray(data, dtype=dtype, order="C")


@njit("f8[::1](f8[::1], f8[::1], f8, f8)", parallel=True, nogil=True)
def _separate_snow(
    prcp: FloatArray,
    tmin: FloatArray,
    t_rain: np.float64,
    t_snow: np.float64,
) -> FloatArray:
    """Separate snow in precipitation."""
    t_rng = t_rain - t_snow
    snow = np.zeros_like(prcp)

    for t in prange(prcp.shape[0]):
        if tmin[t] > t_rain:
            snow[t] = 0.0
        elif tmin[t] < t_snow:
            snow[t] = prcp[t]
        else:
            snow[t] = prcp[t] * (t_rain - tmin[t]) / t_rng
    return snow


def separate_snow(clm: xr.Dataset, t_rain: float, t_snow: float) -> xr.Dataset:
    r"""Separate snow from precipitation.

    Notes
    -----
    The snow separation is based on the Martinez and Gupta (2010) method.
    For more details please refer to Equation 3 in the following publication:
    https://agupubs.onlinelibrary.wiley.com/doi/10.1029/2009WR008294

    Also, for the parameters of this model, please refer to the following publication:
    https://link.springer.com/article/10.1007/s00376-012-2192-7

    The equation is as follows:

    .. math::

        \text{Snow}(t) =
        \begin{cases}
        0, & T^{\text{rain}} < T^{\min}(t) \\
        P(t) \times \dfrac{T^{\text{rain}} - T^{\min}(t)}{T^{\text{rain}} - T^{\text{snow}}}, & T^{\text{snow}} < T^{\min}(t) < T^{\text{rain}} \\
        P(t), & T^{\min}(t) < T^{\text{snow}}
        \end{cases}

    Parameters
    ----------
    clm : xarray.Dataset
        Dataset containing the following variables:

        - ``prcp``: precipitation (mm/day) with time, x, and y dims
        - ``tmin``: minimum temperature (°C) with time, x, and y dims

    t_rain : float
        Rain temperature threshold (°C).
    t_snow : float
        Snow temperature threshold (°C).

    Returns
    -------
    xarray.Dataset
        Dataset containing the following additional variable:

        - ``snow``: daily snowfall (mm)
    """

    def snow_func(
        prcp: FloatArray,
        tmin: FloatArray,
        t_rain: float,
        t_snow: float,
    ) -> FloatArray:
        """Separate snow based on Martinez and Gupta (2010)."""
        return _separate_snow(
            to_numpy(prcp),
            to_numpy(tmin),
            np.float64(t_rain),
            np.float64(t_snow),
        )

    clm["snow"] = xr.apply_ufunc(
        snow_func,
        clm["prcp"],
        clm["tmin"],
        kwargs={"t_rain": t_rain, "t_snow": t_snow},
        input_core_dims=[["time"], ["time"]],
        output_core_dims=[["time"]],
        vectorize=True,
        output_dtypes=[clm["prcp"].dtype],
    ).transpose("time", "y", "x")
    clm["snow"].attrs["units"] = "mm"
    clm["snow"].attrs["long_name"] = "Daily Snowfall"
    return clm



"""Main functions of hysetter."""

from __future__ import annotations

import functools
import shutil
from dataclasses import dataclass
from datetime import datetime  # noqa: TC003
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Literal, overload

import yaml
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    ValidationError,
    field_validator,
    model_validator,
)
from pydantic_core import core_schema
from typing_extensions import Self, SupportsIndex

from hysetter.aoi import get_aoi
from hysetter.forcing import get_forcing
from hysetter.nid import get_nid
from hysetter.nlcd import get_nlcd
from hysetter.nwis import get_streamflow
from hysetter.rasters import get_rasters
from hysetter.soil import get_soil
from hysetter.topo import get_topo

if TYPE_CHECKING:
    from collections.abc import Iterable

    from pydantic_core.core_schema import ValidationInfo
    from yaml.nodes import Node

__all__ = [
    "Config",
    "read_config",
    "write_config",
]

yaml_load = functools.partial(yaml.load, Loader=getattr(yaml, "CSafeLoader", yaml.SafeLoader))
SafeDumper = getattr(yaml, "CSafeDumper", yaml.SafeDumper)


class _PathDumper(SafeDumper):  # pyright: ignore[reportGeneralTypeIssues,reportUntypedBaseClass]
    """A dumper that can represent pathlib.Path objects as strings."""

    def represent_data(self, data: Any) -> Node:
        """Represent Path objects as strings."""
        if isinstance(data, Path):
            return self.represent_scalar("tag:yaml.org,2002:str", str(data))
        return super().represent_data(data)


def _yaml_dump(o: Any, **kwargs: Any) -> str:
    """Dump YAML.

    Notes
    -----
    When python/mypy#1484 is solved, this can be ``functools.partial``
    """
    return yaml.dump(
        o,
        Dumper=_PathDumper,
        stream=None,
        default_flow_style=False,
        indent=2,
        sort_keys=False,
        **kwargs,
    )


class Project(BaseModel):
    """Project information.

    Parameters
    ----------
    name : str
        Name of the project.
    data_dir : str
        Path to the directory where the data will be stored.
    """

    name: str
    data_dir: str


class AOI(BaseModel):
    """Area of interest.

    Notes
    -----
    Only one of ``huc_ids``, ``nhdv2_ids``, ``gagesii_basins``, or
    ``geometry_file`` must be provided.

    Parameters
    ----------
    huc_ids : list, optional
        List of HUC IDs, by default ``None``. The IDs must be strings and HUC
        level will be determined by the length of the string (even numbers
        between 2 and 12).
    nhdv2_ids : list, optional
        List of NHD Feature IDs, by default ``None``. The IDs must be integers
        and are assumed to be NHDPlus V2 catchment IDs.
    gagesii_basins : list, optional
        List of GAGES-II basin IDs, by default ``None``. The IDs must be strings.
    mainstem_main : int, optional
        NHDPlus V2 mainstem ID to get only its main upstream flowlines,
        by default ``None``. The ID must be an integer.
    mainstem_tributaries : bool, optional
        NHDPus V2 mainstem ID to get all its upstream tributaries, by default ``None``.
        The ID must be an integer.
    geometry_file : str, optional
        Path to a geometry file, by default ``None``. Supported file extensions are
        ``.feather``, ``.parquet``, and any format supported by
        ``geopandas.read_file`` (e.g., ``.shp``, ``.geojson``, and ``.gpkg``).
    nhdv2_flowlines : bool, optional
        Whether to retrieve the NHDPlus V2 flowlines within the AOI, by default False.
    streamcat_attrs : list, optional
        StreamCat attributes to retrieve, by default ``None``. This will
        set the ``nhdv2_flowlines`` to True. Use ``pynhd.StreamCat().metrics_df``
        to get a dataframe of all available attributes with their descriptions.
    nldi_attrs : list, optional
        List of slelect attributes to retrieve, by default ``None``.
        Use ``pynhd.nhdplus_attrs_s3()`` to get a dataframe of all available
        attributes with their descriptions.
    """

    huc_ids: list[str] | None = None
    nhdv2_ids: list[int] | None = None
    gagesii_basins: list[str] | None = None
    mainstem_main: int | None = None
    mainstem_tributaries: int | None = None
    geometry_file: str | None = None
    nhdv2_flowlines: bool = False
    streamcat_attrs: list[str] | None = None
    nldi_attrs: list[str] | None = None

    @model_validator(mode="after")
    def _check_exclusive_options(self) -> Self:
        """Check if only one of the options is provided."""
        provided_options = [
            option
            for option in (
                self.huc_ids,
                self.nhdv2_ids,
                self.gagesii_basins,
                self.mainstem_main,
                self.geometry_file,
            )
            if option
        ]
        if len(provided_options) != 1:
            options = (
                "`huc_ids`",
                "`nhdv2_ids`",
                "`gagesii_basins`",
                "`mainstem_main`",
                "`geometry_file`",
            )
            op_list = ", ".join(options[:-1]) + f", or {options[-1]}"
            raise ValueError(f"Only one of {op_list} must be provided.")
        return self


class Forcing(BaseModel):
    """Forcing data.

    Parameters
    ----------
    source : str
        Source of the forcing data. Supported sources are "daymet", "gridmet", and "nldas2".
    start_date : datetime
        Start date of the forcing data.
    end_date : datetime
        End date of the forcing data.
    variables : list, optional
        List of variables to retrieve, by default ``None``. If not provided, all available
        variables will be retrieved.
    crop : bool, optional
        Whether to crop the data to the geometry of the AOI, by default True.
        If False, the data will be saved for the bounding box of the AOI.
    geometry_buffer : int, optional
        Buffer distance in meters to add to the geometry of the AOI before requesting
        the data, by default 0. This is useful for cases where additional post-processing
        is needed that are sensitive to the edges of the data.
    """

    source: Literal["daymet", "gridmet", "nldas2"]
    start_date: datetime
    end_date: datetime
    variables: list[str] | None = None
    crop: bool = True
    geometry_buffer: Annotated[float, Field(strict=True, ge=0)] = 0


class Topo(BaseModel):
    """Topographic data.

    Parameters
    ----------
    resolution_m : int
        Resolution of the data in meters.
    derived_variables : list, optional
        List of derived variables to calculate, by default ``None``. Supported derived
        variables are "slope", "aspect", and "curvature".
    crop : bool, optional
        Whether to crop the data to the geometry of the AOI, by default True.
        If False, the data will be saved for the bounding box of the AOI.
    geometry_buffer : int, optional
        Buffer distance in meters to add to the geometry of the AOI before requesting
        the data, by default 0. This is useful for cases where additional post-processing
        is needed that are sensitive to the edges of the data.
    """

    resolution_m: int
    derived_variables: list[Literal["slope", "aspect", "curvature"]] | None = None
    crop: bool = True
    geometry_buffer: Annotated[float, Field(strict=True, ge=0)] = 0


class RemoteRasters(BaseModel):
    """Remote raster data configuration.

    Parameters
    ----------
    crop : bool, optional
        Whether to crop the data to the geometry of the AOI, by default True.
    geometry_buffer : float, optional
        Buffer distance in meters to add to the AOI geometry before requesting data.

    Notes
    -----
    Additional fields are treated as raster names (URLs). Raster names are sanitized:
        - Lowercased
        - Stripped of whitespace
        - Spaces replaced with underscores
    """

    crop: bool = True
    geometry_buffer: Annotated[float, Field(strict=True, ge=0)] = 0
    model_config = ConfigDict(extra="allow", from_attributes=True)

    @field_validator("*", mode="before")
    @classmethod
    def sanitize_keys(cls, v: Any, info: ValidationInfo) -> Any:
        """Sanitize raster names (keys) and validate URL strings."""
        if info.field_name not in ("crop", "geometry_buffer"):
            if not isinstance(v, str):
                raise ValueError(f"URL must be a string, got {type(v)}")
            return v.strip()
        return v

    @property
    def rasters(self) -> dict[str, str]:
        """Return a dictionary of raster names and URLs."""
        return {k: v for k, v in self.model_dump().items() if k not in ("crop", "geometry_buffer")}


class Soil(BaseModel):
    """Soil data.

    Parameters
    ----------
    source : str
        Source of the soil data. Supported sources are``soilgrids``, ``gnatsgo``,
        and ``polaris``.
    variables : list
        List of variables to retrieve. Each source has its own set of variables.
    crop : bool, optional
        Whether to crop the data to the geometry of the AOI, by default True.
        If False, the data will be saved for the bounding box of the AOI.
    geometry_buffer : int, optional
        Buffer distance in meters to add to the geometry of the AOI before requesting
        the data, by default 0. This is useful for cases where additional post-processing
        is needed that are sensitive to the edges of the data.
    """

    source: Literal["soilgrids", "gnatsgo", "polaris"]
    variables: list[str]
    crop: bool = True
    geometry_buffer: Annotated[float, Field(strict=True, ge=0)] = 0


class NLCD(BaseModel):
    """National Land Cover Database (NLCD) data.

    Parameters
    ----------
    cover : list, optional
        List of years for land cover data, by default ``None``., which defaults to
        the most recent data. Available years are 2021, 2019, 2016, 2013, 2011,
        2008, 2006, 2004, and 2001.
    impervious : list, optional
        List of years for impervious data, by default ``None``., which defaults to
        the most recent data. Available years are 2021, 2019, 2016, 2013, 2011,
        2008, 2006, 2004, and 2001.
    canopy : list, optional
        List of years for canopy data, by default ``None``., which defaults to
        the most recent data. Available years are between 2011 and 2022.
    descriptor : list, optional
        List of years for descriptor data, by default ``None``., which defaults to
        the most recent data. Available years are 2021, 2019, 2016, 2013, 2011,
        2008, 2006, 2004, and 2001.
    crop : bool, optional
        Whether to crop the data to the geometry of the AOI, by default True.
        If False, the data will be saved for the bounding box of the AOI.
    geometry_buffer : int, optional
        Buffer distance in meters to add to the geometry of the AOI before requesting
        the data, by default 0. This is useful for cases where additional post-processing
        is needed that are sensitive to the edges of the data.
    """

    cover: list[int] | None = None
    impervious: list[int] | None = None
    canopy: list[int] | None = None
    descriptor: list[int] | None = None
    crop: bool = True
    geometry_buffer: Annotated[float, Field(strict=True, ge=0)] = 0


class NID(BaseModel):
    """National Inventory of Dams (NID) data.

    Parameters
    ----------
    within_aoi : bool
        Whether to retrieve the NID data within the AOI.
    """

    within_aoi: bool


class Streamflow(BaseModel):
    """Streamflow data from NWIS.

    Parameters
    ----------
    start_date : datetime
        Start date of the streamflow data.
    end_date : datetime
        End date of the streamflow data.
    frequency : str
        Frequency of the streamflow data. Supported frequencies are
        "daily" and "instantaneous".
    within_aoi : bool, optional
        Whether to retrieve the streamflow data for USGS stations that
        are within the AOI, by default True.
    use_col : str, optional
        Instead of getting data for all stations within the AOI, get data
        for a column in the AOI dataframe that contains the USGS station
        IDs, by default None.
    """

    start_date: datetime
    end_date: datetime
    frequency: Literal["daily", "instantaneous"]
    within_aoi: bool = True
    use_col: str | None = None

    @model_validator(mode="after")
    def _check_exclusive_options(self) -> Self:
        """Check if the frequency is either 'daily' or 'instantaneous'."""
        if self.frequency not in ("daily", "instantaneous"):
            raise ValueError("Frequency must be either 'daily' or 'instantaneous'.")
        if self.use_col:
            self.within_aoi = False
        return self


class FileList(list[Path]):
    """A list of files in a directory, ensuring all elements are Path objects within the directory."""

    def __init__(self, parent: str | Path) -> None:
        """
        Initialize the FileList object.

        Parameters
        ----------
        parent : str | Path
            The root directory containing the files.
        """
        self.parent = Path(parent).resolve()
        super().__init__()

    def append(self, file: str | Path) -> None:
        """Append a file as a Path relative to the parent directory."""
        super().append(self.parent / file)

    def extend(self, files: Iterable[str | Path]) -> None:
        """Extend the list with multiple files as Paths relative to the parent directory."""
        super().extend(self.parent / f for f in files)

    def mkdir(self, exist_ok: bool = True, parents: bool = True) -> None:
        """Create the parent directory if it does not exist."""
        self.parent.mkdir(exist_ok=exist_ok, parents=parents)

    def rm_tree(self) -> None:
        """Remove the directory and all its contents."""
        if self.parent.exists():
            shutil.rmtree(self.parent)

    def __get_pydantic_core_schema__(self, handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        """Define Pydantic schema for serialization and validation."""
        return core_schema.no_info_plain_validator_function(FileList._validate)

    @classmethod
    def _validate(cls, value: Any) -> FileList:
        """Validate the value as a FileList object."""
        if isinstance(value, FileList):
            return value
        elif isinstance(value, list):
            return cls(Path.cwd())
        raise TypeError(f"Cannot validate {value} as FileList")

    @overload
    def __getitem__(self, key: SupportsIndex, /) -> Path: ...

    @overload
    def __getitem__(self, key: slice, /) -> FileList: ...

    def __getitem__(self, key: SupportsIndex | slice, /) -> Path | FileList:
        """Retrieve an item or a slice."""
        if isinstance(key, slice):
            return FileList(self.parent, super().__getitem__(key))  # pyright: ignore[reportCallIssue]
        return super().__getitem__(key)

    @overload
    def __setitem__(self, key: SupportsIndex, value: str | Path, /) -> None: ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[str | Path], /) -> None: ...

    def __setitem__(
        self, key: SupportsIndex | slice, value: str | Path | Iterable[str | Path], /
    ) -> None:
        """Allow assignment by index, expanding the list if necessary."""
        if isinstance(key, slice):
            super().__setitem__(key, [self.parent / f for f in value])  # pyright: ignore[reportGeneralTypeIssues]
        elif isinstance(key, int):
            if key >= len(self):
                self.extend([""] * (key - len(self) + 1))  # Expand the list
            super().__setitem__(key, self.parent / value)  # pyright: ignore[reportOperatorIssue]
        else:
            raise TypeError("Index must be an integer or a slice.")

    def __repr__(self) -> str:
        """Return a string representation of the FileList object."""
        files = "\n".join(str(f) for f in self)
        return f"FileList({self.parent}, {len(self)} files):\n{files}"


@dataclass
class FilePaths:
    """File paths to store the data."""

    project: Path
    aoi_parquet: Path
    flowlines: FileList
    streamcat_attrs: FileList
    nldi_attrs: FileList
    forcing: FileList
    topo: FileList
    soil: FileList
    nlcd: FileList
    nid: FileList
    streamflow: FileList
    remote_rasters: dict[str, FileList]

    def rm_tree(self) -> None:
        """Remove the directory and all its contents."""
        shutil.rmtree(self.project)


class Config(BaseModel):
    """Configuration for HySetter.

    Notes
    -----
    Only ``project`` and ``aoi`` are required. The rest are optional.

    Parameters
    ----------
    project : Project
        Project information.
    aoi : AOI
        Area of interest.
    forcing : Forcing, optional
        Forcing data, by default ``None``.
    topo : Topo, optional
        Topographic data, by default ``None``.
    soil : Soil, optional
        Soil data, by default ``None``.
    nlcd : NLCD, optional
        National Land Cover Database (NLCD) data, by default ``None``.
    nid : NID, optional
        National Inventory of Dams (NID) data, by default ``None``.
    streamflow : Streamflow, optional
        Streamflow data from NWIS, by default ``None``.
    remote_rasters : RemoteRasters, optional
        Remote raster data, by default ``None``.
    overwrite : bool, optional
        Whether to overwrite existing data, by default ``False``.
    """

    project: Project
    aoi: AOI
    forcing: Forcing | None = None
    topo: Topo | None = None
    soil: Soil | None = None
    nlcd: NLCD | None = None
    nid: NID | None = None
    streamflow: Streamflow | None = None
    remote_rasters: RemoteRasters | None = None
    overwrite: bool = False
    file_paths: FilePaths = Field(default=None, init=False)  # pyright: ignore[reportAssignmentType]

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        project_dir = Path(self.project.data_dir, self.project.name.replace(" ", "_"))
        project_dir.mkdir(exist_ok=True, parents=True)
        raster_names = list(self.remote_rasters.rasters) if self.remote_rasters else ["rasters"]
        self.file_paths = FilePaths(
            project=project_dir,
            aoi_parquet=Path(project_dir, "aoi.parquet"),
            flowlines=FileList(Path(project_dir, "nhdv2_flowlines")),
            streamcat_attrs=FileList(Path(project_dir, "streamcat_attrs")),
            nldi_attrs=FileList(Path(project_dir, "nldi_attrs")),
            forcing=FileList(Path(project_dir, "forcing")),
            topo=FileList(Path(project_dir, "topo")),
            soil=FileList(Path(project_dir, "soil")),
            nlcd=FileList(Path(project_dir, "nlcd")),
            nid=FileList(Path(project_dir, "nid")),
            streamflow=FileList(Path(project_dir, "streamflow")),
            remote_rasters={n: FileList(Path(project_dir, n)) for n in raster_names},
        )
        if self.overwrite:
            self.file_paths.rm_tree()

    def get_data(self) -> None:
        """Iterate over non-None attributes."""
        get_aoi(self.aoi, self)
        if self.forcing:
            get_forcing(self.forcing, self)
        if self.topo:
            get_topo(self.topo, self)
        if self.soil:
            get_soil(self.soil, self)
        if self.nlcd:
            get_nlcd(self.nlcd, self)
        if self.nid:
            get_nid(self.nid, self)
        if self.streamflow:
            get_streamflow(self.streamflow, self)
        if self.remote_rasters:
            get_rasters(self.remote_rasters, self)


def read_config(file_path: str | Path) -> Config:
    """Read a configuration file and return a Config object.

    Parameters
    ----------
    file_path : str or pathlib.Path
        Path to the configuration file.

    Returns
    -------
    Config
        A Config object.
    """
    config_data = yaml_load(Path(file_path).read_text())
    try:
        return Config(**config_data)
    except ValidationError as e:
        raise ValueError(e) from e


def write_config(config: Config, file_path: str | Path) -> None:
    """Write a Config object to a file.

    Parameters
    ----------
    config : Config
        A Config object.
    file_path : str or pathlib.Path
        Path to the configuration file.
    """
    Path(file_path).write_text(_yaml_dump(config.model_dump()))



"""Utility functions for printing version information.

The original script is from
`xarray <https://github.com/pydata/xarray/blob/main/xarray/util/print_versions.py>`__
"""

from __future__ import annotations

import contextlib
import locale
import os
import platform
import struct
import subprocess
import sys
from importlib.metadata import PackageNotFoundError, distribution
from importlib.metadata import version as get_version
from pathlib import Path
from typing import TextIO

__all__ = ["show_versions"]


def _get_sys_info():
    """Return system information as a dict."""
    blob = []

    # get full commit hash
    commit = None
    if Path(".git").is_dir():
        with contextlib.suppress(Exception):
            pipe = subprocess.Popen(
                ["git", "log", '--format="%H"', "-n", "1"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            so, _ = pipe.communicate()

            if pipe.returncode == 0:
                commit = so
                with contextlib.suppress(ValueError):
                    commit = so.decode("utf-8")
                    commit = commit.strip().strip('"')

    blob.append(("commit", commit))

    with contextlib.suppress(Exception):
        (sysname, _, release, _, machine, processor) = platform.uname()
        blob.extend(
            [
                ("python", sys.version),
                ("python-bits", struct.calcsize("P") * 8),
                ("OS", str(sysname)),
                ("OS-release", str(release)),
                ("machine", str(machine)),
                ("processor", str(processor)),
                ("byteorder", str(sys.byteorder)),
                ("LC_ALL", os.environ.get("LC_ALL", "None")),
                ("LANG", os.environ.get("LANG", "None")),
                ("LOCALE", str(locale.getlocale())),
            ]
        )
    return blob


def _get_package_version(modname: str) -> str:
    try:
        _ = distribution(modname)
        try:
            return get_version(modname)
        except (NotImplementedError, AttributeError):
            return "installed"
    except PackageNotFoundError:
        return "N/A"


def show_versions(file: TextIO = sys.stdout) -> None:
    """Print versions of all the dependencies.

    Parameters
    ----------
    file : file-like, optional
        print to the given file-like object. Defaults to sys.stdout.
    """
    deps = [
        # HyRiver packages
        "async-retriever",
        "pygeoogc",
        "pygeoutils",
        "pynhd",
        "pygridmet",
        "pydaymet",
        "hydrosignatures",
        "pynldas2",
        "pygeohydro",
        "seamless-3dep",
        "tiny-retriever",
        # performance
        "flox",
        "numba",
        "numbagg",
        "planetary-computer",
        "py7zr",
        "pydantic",
        "pyogrio",
        "pystac-client",
        "pyyaml",
        "rich-click",
    ]
    deps_blob = {modname: _get_package_version(modname) for modname in deps}

    print("\nSYS INFO", file=file)
    print("--------", file=file)
    for k, stat in _get_sys_info():
        print(f"{k}: {stat}", file=file)

    pad = len(max(deps, key=len)) + 1
    header = f"\n{'PACKAGE':<{pad}}  VERSION"
    print(header, file=file)
    print("-" * len(header), file=file)
    for k, stat in deps_blob.items():
        print(f"{k:<{pad}}  {stat}", file=file)
    print("-" * len(header), file=file)



"""Main functions of hysetter."""

from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
from rich.console import Console
from rich.progress import track

if TYPE_CHECKING:
    from pathlib import Path

    from pyproj import CRS
    from shapely import Polygon
    from xarray import DataArray

    from hysetter.hysetter import Config, Topo

__all__ = ["get_topo"]


def _slope(dem: DataArray) -> DataArray:
    """Calculate slope in degrees."""
    from xrspatial import slope

    return slope(dem)


def _aspect(dem: DataArray) -> DataArray:
    """Calculate aspect in degrees.

    It is measured clockwise in degrees with 0 (due north), and 360 (again due north).
    """
    from xrspatial import aspect

    return aspect(dem)


def _curvature(dem: DataArray) -> DataArray:
    """Calculate curvature.

    A positive curvature indicates the surface is upwardly convex.
    A negative value indicates it is upwardly concave. A value of 0 indicates a flat surface.
    """
    from xrspatial import curvature

    return curvature(dem)


def _read_tiffs(tiff_files: list[Path], poly: Polygon, crs: CRS) -> DataArray:
    """Convert a list of tiff files to a vrt file and return a xarray.DataArray."""
    from seamless_3dep import tiffs_to_da

    topo = tiffs_to_da(tiff_files, poly, crs)
    topo.name = "elevation"
    return topo


def get_topo(data_cfg: Topo, model_cfg: Config) -> None:
    """Get topo data for the area of interest."""
    import xarray as xr
    from seamless_3dep import get_map
    from shapely import box

    console = Console(force_jupyter=False)

    topo_functions = {"slope": _slope, "aspect": _aspect, "curvature": _curvature}

    gdf = gpd.read_parquet(model_cfg.file_paths.aoi_parquet)
    topo_paths = model_cfg.file_paths.topo
    topo_paths.mkdir()
    if data_cfg.geometry_buffer > 0:
        buffer = data_cfg.geometry_buffer
    else:
        buffer = 10 * data_cfg.resolution_m
    join_style = "round"
    if not data_cfg.crop:
        gdf["geometry"] = box(*gdf.geometry.bounds.to_numpy().T)
        join_style = "mitre"
    gdf = gdf.to_crs(5070).buffer(buffer, join_style=join_style).to_crs(4326)
    for i, geom in track(
        enumerate(gdf.geometry),
        description="Getting DEM from 3DEP",
        total=len(gdf),
        console=console,
    ):
        topo_paths[i] = f"topo_geom_{i}.nc"
        if topo_paths[i].exists():
            continue
        try:
            tiff_files = get_map("DEM", geom.bounds, topo_paths.parent, data_cfg.resolution_m)
            topo = _read_tiffs(tiff_files, geom, gdf.crs)  # pyright: ignore[reportArgumentType]
        except Exception:
            console.print_exception(show_locals=True, max_frames=4)
            console.print(f"Failed to get topo data for AOI index {i}")
            continue
        if data_cfg.derived_variables:
            topo.attrs["res"] = topo.rio.resolution()
            topo_list = [
                topo_functions[var](topo)
                for var in data_cfg.derived_variables
                if var in topo_functions
            ]
            topo = xr.merge([topo, *topo_list])
        topo.to_netcdf(topo_paths[i])



"""Main functions of hysetter."""

from __future__ import annotations

from typing import TYPE_CHECKING

import geopandas as gpd
from rich.console import Console
from rich.progress import track

if TYPE_CHECKING:
    from hysetter.hysetter import Config, RemoteRasters

__all__ = ["get_rasters"]


def get_rasters(data_cfg: RemoteRasters, model_cfg: Config) -> None:
    """Get raster data for the area of interest."""
    import rioxarray as rxr
    from shapely import box

    console = Console(force_jupyter=False)
    gdf = gpd.read_parquet(model_cfg.file_paths.aoi_parquet)
    raster_paths = model_cfg.file_paths.remote_rasters
    join_style = "round"
    if not data_cfg.crop:
        gdf["geometry"] = box(*gdf.geometry.bounds.to_numpy().T)
        join_style = "mitre"
    if data_cfg.geometry_buffer > 0:
        gdf = gdf.to_crs(5070).buffer(data_cfg.geometry_buffer, join_style=join_style)
    for name, url in data_cfg.rasters.items():
        raster_paths[name].mkdir()
        for i, geom in track(
            enumerate(gdf.geometry),
            description=f"Getting raster data for {name}",
            total=len(gdf),
            console=console,
        ):
            raster_paths[name][i] = f"{name}_geom_{i}.tif"
            if raster_paths[name][i].exists():
                continue
            try:
                rxr.open_rasterio(url).squeeze(drop=True).rio.clip_box(
                    *geom.bounds, crs=gdf.crs, auto_expand=True
                ).rio.clip([geom], crs=gdf.crs, all_touched=True).rio.to_raster(
                    raster_paths[name][i]
                )
            except Exception:
                console.print_exception(show_locals=True, max_frames=4)
                console.print(f"Failed to get raster data from {url} for AOI index {i}")
                continue



"""Nox sessions."""

from __future__ import annotations

import shutil
from pathlib import Path

import nox

try:
    import tomllib as tomli
except ImportError:
    import tomli


def get_package_name() -> str:
    """Get the name of the package."""
    with Path("pyproject.toml").open("rb") as f:
        return tomli.load(f)["project"]["name"]


def get_extras() -> list[str]:
    """Get the name of the package."""
    with Path("pyproject.toml").open("rb") as f:
        extras = tomli.load(f)["project"]["optional-dependencies"]
    return [e for e in extras if e not in ("test", "typeguard")]


def get_deps() -> list[str]:
    """Get the name of the package."""
    with Path("pyproject.toml").open("rb") as f:
        return tomli.load(f)["project"]["dependencies"]


python_versions = ["3.10"]
lint_versions = ["3.11"]
package = get_package_name()
gh_deps = {
    "async-retriever": [],
    "hydrosignatures": [],
    "pygeoogc": ["async-retriever"],
    "pygeoutils": ["async-retriever", "pygeoogc"],
    "pynhd": ["async-retriever", "pygeoogc", "pygeoutils"],
    "py3dep": ["async-retriever", "pygeoogc", "pygeoutils"],
    "pygeohydro": ["async-retriever", "pygeoogc", "pygeoutils", "pynhd", "hydrosignatures"],
    "pydaymet": ["async-retriever", "pygeoogc", "pygeoutils", "py3dep"],
    "pygridmet": ["async-retriever", "pygeoogc", "pygeoutils"],
    "pynldas2": ["async-retriever", "pygeoutils"],
    "hysetter": [
        "async-retriever",
        "pygeoogc",
        "pygeoutils",
        "pynhd",
        "hydrosignatures",
        "pygeohydro",
        "py3dep",
        "pydaymet",
        "pygridmet",
        "pynldas2",
    ],
}
nox.options.sessions = (
    "pre-commit",
    "type-check",
    "tests",
)


def install_deps(
    session: nox.Session, extra: str | None = None, version_limit: list[str] | None = None
) -> None:
    """Install package dependencies."""
    deps = [f".[{extra}]"] if extra else ["."]
    deps += [f"git+https://github.com/hyriver/{p}.git" for p in gh_deps[package]]
    if version_limit:
        deps += list(version_limit)
    session.install(*deps)
    dirs = [".pytest_cache", "build", "dist", ".eggs"]
    for d in dirs:
        shutil.rmtree(d, ignore_errors=True)

    patterns = ["*.egg-info", "*.egg", "*.pyc", "*~", "**/__pycache__"]
    for p in patterns:
        for f in Path.cwd().rglob(p):
            shutil.rmtree(f, ignore_errors=True)


@nox.session(name="pre-commit", python=lint_versions)
def pre_commit(session: nox.Session) -> None:
    """Lint using pre-commit."""
    session.install("pre-commit")
    session.run(
        "pre-commit",
        "run",
        "--all-files",
        "--hook-stage=manual",
        *session.posargs,
    )


@nox.session(name="type-check", python=python_versions)
def type_check(session: nox.Session) -> None:
    """Run Pyright."""
    extras = get_extras()
    install_deps(session, ",".join(extras))
    session.install("pyright")
    session.run("pyright")


@nox.session(python=python_versions)
def tests(session: nox.Session) -> None:
    """Run the test suite."""
    extras = get_extras()
    speedup_dep = True
    try:
        extras.remove("speedup")
    except ValueError:
        speedup_dep = False

    install_deps(session, ",".join(["test", *extras]))
    session.run("pytest", "--cov", "--cov-append", "--cov-report=xml", *session.posargs)
    session.notify("cover")
    if speedup_dep:
        session.notify("speedup")


@nox.session(python=python_versions)
def speedup(session: nox.Session) -> None:
    """Run tests that require speedup deps."""
    extras = get_extras()
    install_deps(session, ",".join(["test", *extras]))
    session.run("pytest", "-m", "speedup", *session.posargs)


@nox.session
def cover(session: nox.Session) -> None:
    """Coverage analysis."""
    session.install("coverage[toml]")
    session.run("coverage", "report")
    session.run("coverage", "html")



"""Main functions of hysetter."""

from __future__ import annotations

import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Literal, cast

import geopandas as gpd
import pandas as pd
from rich.console import Console
from rich.progress import track

from hysetter.exceptions import InputTypeError

if TYPE_CHECKING:
    from hysetter.hysetter import AOI, Config, FileList

__all__ = ["get_aoi"]


def _get_hucs(huc_ids: list[str]) -> gpd.GeoDataFrame:
    """Get HUCs."""
    from pynhd import WaterData

    huc_df = pd.Series(huc_ids).astype(str)
    if not huc_df.str.len().isin([2, 4, 6, 8, 10, 12]).all():
        raise InputTypeError("huc_ids", "strings of lengths between 2 and 12.")
    huc_dict = huc_df.groupby(huc_df.str.len()).apply(list).to_dict()
    aoi_list = []
    for lvl, ids in huc_dict.items():
        wd = WaterData(f"wbd{str(lvl).zfill(2)}")  # pyright: ignore[reportArgumentType]
        aoi_list.append(wd.byid(f"huc{lvl}", ids))
    return gpd.GeoDataFrame(pd.concat(aoi_list, ignore_index=True))


def _get_mainstem(
    mainstem_id: int,
    navigation: Literal["upstreamMain", "upstreamTributaries"],
    get_flw: bool,
    flw_dir: FileList,
    sc_attrs: list[str] | None,
    sc_dir: FileList,
    nldi_attrs: list[str] | None,
    nldi_dir: FileList,
) -> gpd.GeoDataFrame:
    """Get NHDPlus V2 catchments for ComIDs of a mainstem."""
    import pygeoutils as pgu
    from pynhd import NLDI, GeoConnex, WaterData, streamcat

    outlet_id = (
        GeoConnex("mainstems")
        .byid("id", str(mainstem_id))
        .outlet_nhdpv2_comid.str.split("/")
        .str[-1]
        .iloc[0]
    )
    nldi = NLDI()
    comids = nldi.navigate_byid(
        "comid", outlet_id, navigation, source="flowlines", distance=9999
    ).nhdplus_comid.tolist()
    if get_flw:
        wd = WaterData("nhdflowline_network")
        flw_dir.mkdir()
        flw_dir[0] = "flowlines.parquet"
        if not flw_dir[0].exists():
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                flw = wd.byid("comid", comids)
                flw.to_parquet(flw_dir[0])
    if sc_attrs:
        sc_dir.mkdir()
        sc_dir[0] = "streamcat_attrs.parquet"
        if not sc_dir[0].exists():
            streamcat(sc_attrs, "cat", comids).to_parquet(sc_dir[0])

    if nldi_attrs:
        nldi_dir.mkdir()
        nldi_dir[0] = "nldi_attrs.parquet"
        if not nldi_dir[0].exists():
            nldi.get_characteristics(nldi_attrs, comids).to_parquet(nldi_dir[0])

    wd = WaterData("catchmentsp")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        cat = wd.byid("featureid", comids)
        return pgu.multi2poly(cat)


def _read_geometry_file(geometry_file: str) -> gpd.GeoDataFrame:
    """Read geometry file."""
    file_ext = Path(geometry_file).suffix
    if file_ext == ".parquet":
        gdf = pd.read_parquet(geometry_file)
    elif file_ext == ".feather":
        gdf = pd.read_feather(geometry_file)
    else:
        gdf = gpd.read_file(geometry_file, engine="pyogrio")
    if not gdf.geom_type.isin(["Polygon", "MultiPolygon"]).all():
        raise ValueError("Geometry file must contain polygons or multipolygons.")
    gdf = gdf.reset_index(drop=True)
    gdf = cast("gpd.GeoDataFrame", gdf)
    return gdf


def _get_flowlines(
    gdf: gpd.GeoDataFrame,
    flw_dir: FileList,
    sc_attrs: list[str] | None,
    sc_dir: FileList,
    nldi_attrs: list[str] | None,
    nldi_dir: FileList,
) -> None:
    """Get flowlines."""
    from pynhd import NLDI, WaterData, streamcat

    console = Console(force_jupyter=False)
    # To avoid instantiating WaterData if data already exists
    # we set it to None here and instantiate it later
    # in the loop if needed
    wd = None

    if sc_attrs or nldi_attrs:
        description = "Getting flowlines and their attributes"
    else:
        description = "Getting flowlines from WaterData"

    for i, geom in track(
        enumerate(gdf.geometry),
        description=description,
        total=len(gdf),
        console=console,
    ):
        # Only instantiate WaterData if not already done
        if wd is None:
            wd = WaterData("nhdflowline_network")
        flw_dir[i] = f"aoi_geom_{i}.parquet"
        if not flw_dir[i].exists():
            flw_dir.mkdir()
            try:
                wd.bygeom(geom, gdf.crs).to_parquet(flw_dir[i])  # pyright: ignore[reportArgumentType]
            except Exception:
                console.print_exception(show_locals=True, max_frames=4)
                console.print(f"Failed to get flowlines for AOI {i}.")

        if sc_attrs:
            sc_dir[i] = f"streamcat_geom_{i}.parquet"
            if not sc_dir[i].exists():
                sc_dir.mkdir()
                comids = gpd.read_parquet(flw_dir[i]).comid.tolist()
                try:
                    streamcat(sc_attrs, "cat", comids).to_parquet(sc_dir[i])
                except Exception:
                    console.print_exception(show_locals=True, max_frames=4)
                    console.print(f"Failed to get StreamCat for AOI {i}.")

        if nldi_attrs:
            nldi_dir[i] = f"nldi_geom_{i}.parquet"
            if not nldi_dir[i].exists():
                nldi_dir.mkdir()
                try:
                    nldi = NLDI()
                    comids = gpd.read_parquet(flw_dir[i]).comid.tolist()
                    nldi.get_characteristics(nldi_attrs, comids).to_parquet(nldi_dir[i])
                except Exception:
                    console.print_exception(show_locals=True, max_frames=4)
                    console.print(f"Failed to get NLDI attrs for AOI {i}.")


def get_aoi(aoi_cfg: AOI, model_cfg: Config) -> None:
    """Get the area of interest."""
    import pygeoutils as geoutils
    from pynhd import WaterData

    console = Console(force_jupyter=False)

    files = model_cfg.file_paths
    if files.aoi_parquet.exists():
        console.print(f"Reading AOI from [bold green]{files.aoi_parquet.resolve()}")
        gdf = gpd.read_parquet(files.aoi_parquet)
    else:
        files.project.mkdir(exist_ok=True, parents=True)
        gdf = None
        if aoi_cfg.huc_ids:
            console.print("Getting AOI: HUCs from WaterData.")
            gdf = _get_hucs(aoi_cfg.huc_ids)
        elif aoi_cfg.nhdv2_ids:
            console.print("Getting AOI: NHDPlusV2 catchments from WaterData.")
            gdf = WaterData("catchmentsp").byid("featureid", aoi_cfg.nhdv2_ids)
            gdf = geoutils.multi2poly(gdf)
        elif aoi_cfg.gagesii_basins:
            console.print("Getting AOI: GAGES-II basins from WaterData.")
            gdf = WaterData("gagesii_basins").byid("gage_id", aoi_cfg.gagesii_basins)
            gdf = geoutils.multi2poly(gdf)
        elif aoi_cfg.mainstem_main:
            console.print("Getting AOI: Mainstem catchments (main only) from GeoConnex.")
            try:
                gdf = _get_mainstem(
                    aoi_cfg.mainstem_main,
                    "upstreamMain",
                    aoi_cfg.nhdv2_flowlines,
                    files.flowlines,
                    aoi_cfg.nldi_attrs,
                    files.nldi_attrs,
                    aoi_cfg.streamcat_attrs,
                    files.streamcat_attrs,
                )
            except Exception:
                console.print_exception(show_locals=True, max_frames=4)
                console.print(f"Failed to get data for {aoi_cfg.mainstem_main}.")
            return
        elif aoi_cfg.mainstem_tributaries:
            console.print("Getting AOI: Mainstem catchments (tributaries) from GeoConnex.")
            try:
                gdf = _get_mainstem(
                    aoi_cfg.mainstem_tributaries,
                    "upstreamTributaries",
                    aoi_cfg.nhdv2_flowlines,
                    files.flowlines,
                    aoi_cfg.nldi_attrs,
                    files.nldi_attrs,
                    aoi_cfg.streamcat_attrs,
                    files.streamcat_attrs,
                )
            except Exception:
                console.print_exception(show_locals=True, max_frames=4)
                console.print(f"Failed to get data for {aoi_cfg.mainstem_main}.")
            return
        elif aoi_cfg.geometry_file:
            console.print(f"Getting AOI: From {aoi_cfg.geometry_file}")
            gdf = _read_geometry_file(aoi_cfg.geometry_file)

        if gdf is None:
            raise ValueError
        gdf.to_parquet(files.aoi_parquet)

    if aoi_cfg.nhdv2_flowlines or aoi_cfg.streamcat_attrs or aoi_cfg.nldi_attrs:
        files.flowlines.mkdir(exist_ok=True, parents=True)
        _get_flowlines(
            gdf,
            files.flowlines,
            aoi_cfg.streamcat_attrs,
            files.streamcat_attrs,
            aoi_cfg.nldi_attrs,
            files.nldi_attrs,
        )



"""Main functions of hysetter."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import geopandas as gpd
from rich.console import Console
from rich.progress import track

if TYPE_CHECKING:
    from hysetter.hysetter import NID, Config

__all__ = ["get_nid"]


def get_nid(nid_cfg: NID, model_config: Config) -> None:
    """Get NID data for the area of interest.

    Parameters
    ----------
    cfg_nid : NID
        A NID object.
    nid_dir : Path
        Path to the directory where the NID data will be saved.
    aoi_parquet : Path
        The path to the AOI parquet file.
    """
    from pygeohydro import NID

    console = Console(force_jupyter=False)
    nid_paths = model_config.file_paths.nid
    nid_paths.mkdir()
    gdf = gpd.read_parquet(model_config.file_paths.aoi_parquet)
    if not nid_cfg.within_aoi:
        return

    nid = None
    for i, geom in track(
        enumerate(gdf.geometry),
        description="Getting dams from NID",
        total=len(gdf),
        console=console,
    ):
        if nid is None:
            nid = NID()
            nid.stage_nid_inventory(Path(nid_paths.parent, "full_nid_inventory.parquet"))
        nid_paths[i] = f"nid_geom_{i}.parquet"
        if nid_paths[i].exists():
            continue
        try:
            nid.get_bygeom(geom, gdf.crs).to_parquet(nid_paths[i])  # pyright: ignore[reportArgumentType]
        except Exception:
            console.print_exception(show_locals=True, max_frames=4)
            console.print(f"Failed to get NID data for AOI index {i}")
            continue
